@@ -1,6 +1,6 @@
 # -*- coding: UTF-8 -*-
 #
-#    Copyright 2008-2010, Lukas Lueg, lukas.lueg@gmail.com
+#    Copyright 2008-2011, Lukas Lueg, lukas.lueg@gmail.com
 #
 #    This file is part of Pyrit.
 #
@@ -29,39 +29,41 @@
 
 from __future__ import with_statement
 
-# prevent call to socket.getfqdn
+from collections import deque
 import BaseHTTPServer
-def fast_address_string(self):
-    return '%s' % self.client_address[0]
-BaseHTTPServer.BaseHTTPRequestHandler.address_string = fast_address_string
-del fast_address_string
-
 import hashlib
-import select
-import SimpleXMLRPCServer
+import random
 import socket
 import sys
-import time
 import threading
+import time
 import uuid
 import util
+import warnings
 import xmlrpclib
 
 import config
 import network
 import storage
+import util
 import _cpyrit_cpu
 
 
+# prevent call to socket.getfqdn
+def fast_address_string(self):
+    return '%s' % self.client_address[0]
+BaseHTTPServer.BaseHTTPRequestHandler.address_string = fast_address_string
+del fast_address_string
+
+
 def version_check(mod):
     ver = getattr(mod, "VERSION", "unknown")
     if ver != _cpyrit_cpu.VERSION:
-        print >> sys.stderr, \
-                "WARNING: Version mismatch between %s ('%s') and %s ('%s')\n" \
-                % (_cpyrit_cpu, _cpyrit_cpu.VERSION, mod, ver)
+        warnings.warn("WARNING: Version mismatch between %s ('%s') and %s " \
+                     "('%s')\n" % (_cpyrit_cpu, _cpyrit_cpu.VERSION, mod, ver))
 
 
-class Core(threading.Thread):
+class Core(util.Thread):
     """The class Core provides basic scheduling and testing. It should not be
        used directly but through sub-classes.
 
@@ -79,7 +81,7 @@ class Core(threading.Thread):
 
     def __init__(self, queue):
         """Create a new Core that pulls work from the given CPyrit instance."""
-        threading.Thread.__init__(self)
+        util.Thread.__init__(self)
         self.queue = queue
         self.compTime = self.resCount = self.callCount = 0
         self.isTested = False
@@ -116,18 +118,17 @@ class Core(threading.Thread):
                 self.compTime += time.time() - t
                 self.resCount += len(res)
                 self.callCount += 1
-                avg = (2 * self.buffersize + (self.resCount / self.compTime * 3)) / 3
-                self.buffersize = int(max(self.minBufferSize,
-                                  min(self.maxBufferSize, avg)))
+                if self.compTime > 0:
+                    # carefully move towards three seconds of execution-time
+                    avg = (2 * self.buffersize + \
+                           (self.resCount / self.compTime * 3)) / 3
+                    self.buffersize = int(max(self.minBufferSize,
+                                              min(self.maxBufferSize, avg)))
                 self.queue._scatter(essid, pwlist, res)
 
     def __str__(self):
         return self.name
 
-    def shutdown(self):
-        self.shallStop = True
-        self.join()
-
 
 class CPUCore(Core, _cpyrit_cpu.CPUDevice):
     """Standard-CPU implementation. The underlying C-code may use VIA Padlock,
@@ -141,12 +142,75 @@ class CPUCore(Core, _cpyrit_cpu.CPUDevice):
         self.start()
 
 
+class LowLatencyCore(Core):
+    def __init__(self, queue):
+        Core.__init__(self, queue)
+        self.bufferSizeDiv = 0
+
+    def _getTestData(self, i):
+        return (Core.TV_ESSID, [Core.TV_PW] * i)
+
+    def _testData(self, res):
+        if any((pmk != Core.TV_PMK for pmk in res)):
+            raise ValueError("Test-vector does not result in correct PMK.")
+
+    def _processData(self, essid, pwlist, res, tm):
+        assert(len(res) == len(pwlist))
+        t = time.time()
+        self.compTime += t - tm
+        self.resCount += len(res)
+        self.callCount += 1
+        avg = (2 * self.buffersize + (self.resCount / self.compTime)) / 3
+        if self.bufferSizeDiv > 0:
+            avg = self.bufferSizeDiv * int((avg + self.bufferSizeDiv - 1) \
+                  / self.bufferSizeDiv)
+        self.buffersize = int(max(self.minBufferSize,
+                              min(self.maxBufferSize, avg)))
+        self.queue._scatter(essid, pwlist, res)
+        return t
+
+    def solve(self, essid, pwlist):
+        enq = self.send(essid, pwlist)
+        assert(enq)
+        return self.receive(True)
+
+    def run(self):
+        work_queue = deque()
+        work_available = False
+        t = time.time()
+        while not self.shallStop:
+            if not work_available:
+                if not self.isTested:
+                    essid, pwlist = self._getTestData(101)
+                else:
+                    essid, pwlist = self.queue._gather(self.buffersize, \
+                                                       timeout=0.5)
+                if essid is not None:
+                    work_queue.append((essid, pwlist, not self.isTested))
+                    self.isTested = True
+                    work_available = True
+                    if len(work_queue) == 1:
+                        t = time.time()
+            if len(work_queue) <= 0:
+                continue
+            if work_available:
+                essid, pwlist, testing = work_queue[-1]
+                work_available = not self.send(essid, pwlist)
+            res = self.receive(work_available)
+            if res is not None:
+                essid, pwlist, testing = work_queue.popleft()
+                if not testing:
+                    t = self._processData(essid, pwlist, res, t)
+                else:
+                    self._testData(res)
+
+
 try:
     import _cpyrit_opencl
 except ImportError:
     pass
 except Exception, e:
-    print >> sys.stderr, "Failed to load Pyrit's OpenCL-driven core ('%s')." % e
+    print >> sys.stderr, "Failed to load Pyrit's OpenCL-core ('%s')." % e
 else:
     version_check(_cpyrit_opencl)
 
@@ -160,7 +224,7 @@ else:
             self.minBufferSize = 1024
             self.buffersize = 4096
             maxhwsize = reduce(lambda x, y: x * y, self.maxWorkSizes)
-            self.maxBufferSize = min(61440, maxhwsize)
+            self.maxBufferSize = min(180224, maxhwsize)
             self.start()
 
 
@@ -169,7 +233,7 @@ try:
 except ImportError:
     pass
 except Exception, e:
-    print >> sys.stderr, "Failed to load Pyrit's CUDA-driven core ('%s')." % e
+    print >> sys.stderr, "Failed to load Pyrit's CUDA-core ('%s')." % e
 else:
     version_check(_cpyrit_cuda)
 
@@ -182,7 +246,29 @@ else:
             self.name = "CUDA-Device #%i '%s'" % (dev_idx + 1, self.deviceName)
             self.minBufferSize = 1024
             self.buffersize = 4096
-            self.maxBufferSize = 40960
+            self.maxBufferSize = 131072
+            self.start()
+
+
+try:
+    import _cpyrit_calpp
+except ImportError:
+    pass
+except Exception, e:
+    print >> sys.stderr, "Failed to load Pyrit's CAL-core ('%s')." % e
+else:
+    version_check(_cpyrit_calpp)
+
+    class CALCore(LowLatencyCore, _cpyrit_calpp.CALDevice):
+        """Computes results on ATI CAL capable devices."""
+
+        def __init__(self, queue, dev_idx):
+            LowLatencyCore.__init__(self, queue)
+            _cpyrit_calpp.CALDevice.__init__(self, dev_idx)
+            self.name = "CAL++ Device #%i '%s'" % \
+                        (dev_idx + 1, self.deviceName)
+            self.minBufferSize, self.buffersize, self.maxBufferSize, \
+                self.bufferSizeDiv = self.workSizes()
             self.start()
 
 
@@ -205,12 +291,12 @@ else:
             self.start()
 
 
-class NetworkCore(Core, SimpleXMLRPCServer.SimpleXMLRPCServer):
+class NetworkCore(util.AsyncXMLRPCServer, Core):
 
-    class NetworkObserver(threading.Thread):
+    class NetworkObserver(util.Thread):
 
         def __init__(self, core):
-            threading.Thread.__init__(self)
+            util.Thread.__init__(self)
             self.core = core
             self.setDaemon(True)
             self.start()
@@ -234,17 +320,15 @@ class NetworkCore(Core, SimpleXMLRPCServer.SimpleXMLRPCServer):
             self.lastseen = time.time()
 
     def __init__(self, queue, host='', port=17935):
-        SimpleXMLRPCServer.SimpleXMLRPCServer.__init__(self, (host, port), \
-                                                        logRequests=False)
+        util.AsyncXMLRPCServer.__init__(self, (host, port))
         Core.__init__(self, queue)
         self.name = "Network-Clients"
         self.uuid = str(uuid.uuid4())
-        self.methods = {'register': self.rpc_register, \
-                        'unregister': self.rpc_unregister, \
-                        'gather': self.rpc_gather, \
-                        'scatter': self.rpc_scatter, \
-                        'revoke': self.rpc_revoke}
-        self.register_instance(self)
+        self.methods['register'] = self.rpc_register
+        self.methods['unregister'] = self.rpc_unregister
+        self.methods['gather'] = self.rpc_gather
+        self.methods['scatter'] = self.rpc_scatter
+        self.methods['revoke'] = self.rpc_revoke
         self.client_lock = threading.Lock()
         self.clients = {}
         self.host = host
@@ -253,18 +337,6 @@ class NetworkCore(Core, SimpleXMLRPCServer.SimpleXMLRPCServer):
         self.startTime = time.time()
         self.start()
 
-    def _dispatch(self, method, params):
-        if method not in self.methods:
-            raise AttributeError
-        else:
-            return self.methods[method](*params)
-
-    def run(self):
-        while not self.shallStop:
-            r, w, e = select.select([self], [], [], 0.5)
-            if r:
-                self.handle_request()
-
     def _get_client(self, uuid):
         with self.client_lock:
             if uuid in self.clients:
@@ -379,6 +451,12 @@ class CPyrit(object):
                         self.cores.append(core)
                         ncpus -= 1
 
+        # CAL++
+        if 'cpyrit._cpyrit_calpp' in sys.modules:
+            for dev_idx, device in enumerate(_cpyrit_calpp.listDevices()):
+                self.cores.append(CALCore(queue=self, dev_idx=dev_idx))
+                ncpus -= 1
+
         #CPUs
         for i in xrange(ncpus):
             self.cores.append(CPUCore(queue=self))
@@ -403,6 +481,8 @@ class CPyrit(object):
                     break
             else:
                 self.ncore_uuid = None
+        else:
+            self.ncore_uuid = None
 
     def _check_cores(self):
         for core in self.cores:
@@ -560,7 +640,8 @@ class CPyrit(object):
                             newslice = pwslice[:restsize]
                             del pwdict[idx]
                             if len(pwslice[len(newslice):]) > 0:
-                                pwdict[idx + len(newslice)] = pwslice[len(newslice):]
+                                pwdict[idx + len(newslice)] = \
+                                    pwslice[len(newslice):]
                             pwslices.append((idx, len(newslice)))
                             passwords.extend(newslice)
                             restsize -= len(newslice)
@@ -634,3 +715,117 @@ class CPyrit(object):
                 d[idx] = passwordlist[ptr:ptr + length]
                 ptr += length
             self.cv.notifyAll()
+
+
+class StorageIterator(object):
+    """Iterates over the database, computes new Pairwise Master Keys if
+       necessary and requested and yields tuples of (password,PMK)-tuples.
+    """
+
+    def __init__(self, storage, essid, \
+                 yieldOldResults=True, yieldNewResults=True):
+        self.cp = CPyrit() if yieldNewResults else None
+        self.workunits = []
+        self.essid = essid
+        self.storage = storage
+        keys = list(self.storage.passwords)
+        random.shuffle(keys)
+        self.iterkeys = iter(keys)
+        self.yieldOldResults = yieldOldResults
+        self.yieldNewResults = yieldNewResults
+
+    def keycount(self):
+        return self.storage.essids.keycount(self.essid)
+
+    def __len__(self):
+        return len(self.storage.passwords)
+
+    def __iter__(self):
+        return self
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        if self.cp is not None:
+            self.cp.shutdown()
+
+    def next(self):
+        while True:
+            try:
+                key = self.iterkeys.next()
+            except StopIteration:
+                if self.yieldNewResults:
+                    solvedPMKs = self.cp.dequeue(block=True)
+                    if solvedPMKs is not None:
+                        solvedEssid, solvedKey, solvedPasswords = \
+                          self.workunits.pop(0)
+                        solvedResults = zip(solvedPasswords, solvedPMKs)
+                        self.storage.essids[solvedEssid, solvedKey] = \
+                          solvedResults
+                        return solvedResults
+                assert len(self.workunits) == 0
+                raise StopIteration
+            else:
+                if self.yieldOldResults:
+                    try:
+                        results = self.storage.essids[self.essid, key]
+                    except KeyError:
+                        pass
+                    else:
+                        return results
+                if self.yieldNewResults:
+                    passwords = self.storage.passwords[key]
+                    self.workunits.append((self.essid, key, passwords))
+                    self.cp.enqueue(self.essid, passwords)
+                    solvedPMKs = self.cp.dequeue(block=False)
+                    if solvedPMKs is not None:
+                        solvedEssid, solvedKey, solvedPasswords = \
+                          self.workunits.pop(0)
+                        solvedResults = zip(solvedPasswords, solvedPMKs)
+                        self.storage.essids[solvedEssid, solvedKey] = \
+                          solvedResults
+                        return solvedResults
+
+
+class PassthroughIterator(object):
+    """A iterator that takes an ESSID and an iterable of passwords, computes
+       the corresponding Pairwise Master Keys and and yields tuples of
+       (password,PMK)-tuples.
+    """
+
+    def __init__(self, essid, iterable, buffersize=20000):
+        self.cp = CPyrit()
+        self.essid = essid
+        self.iterator = iter(iterable)
+        self.workunits = []
+        self.buffersize = buffersize
+
+    def __iter__(self):
+        return self
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.cp.shutdown()
+
+    def next(self):
+        pwbuffer = []
+        for line in self.iterator:
+            pw = line.strip('\r\n')[:63]
+            if len(pw) >= 8:
+                pwbuffer.append(pw)
+            if len(pwbuffer) > self.buffersize:
+                self.workunits.append(pwbuffer)
+                self.cp.enqueue(self.essid, self.workunits[-1])
+                pwbuffer = []
+                solvedPMKs = self.cp.dequeue(block=False)
+                if solvedPMKs is not None:
+                    return zip(self.workunits.pop(0), solvedPMKs)
+        if len(pwbuffer) > 0:
+            self.workunits.append(pwbuffer)
+            self.cp.enqueue(self.essid, self.workunits[-1])
+        for solvedPMKs in self.cp:
+            return zip(self.workunits.pop(0), solvedPMKs)
+        raise StopIteration
