@@ -1,6 +1,6 @@
 # -*- coding: UTF-8 -*-
 #
-#    Copyright 2008-2010, Lukas Lueg, lukas.lueg@gmail.com
+#    Copyright 2008-2011, Lukas Lueg, lukas.lueg@gmail.com
 #
 #    This file is part of Pyrit.
 #
@@ -25,30 +25,96 @@
 
 from __future__ import with_statement
 
-# prevent call to socket.getfqdn
 import BaseHTTPServer
-def fast_address_string(self):
-    return '%s' % self.client_address[0]
-BaseHTTPServer.BaseHTTPRequestHandler.address_string = fast_address_string
-del fast_address_string
-
 import hashlib
 import os
 import random
-import SimpleXMLRPCServer
+import re
 import struct
 import sys
-import zlib
+import threading
 import xmlrpclib
+import zlib
 try:
     import sqlalchemy as sql
 except ImportError:
     pass
 else:
     from sqlalchemy import orm
+    if hasattr(sql, 'LargeBinary'):
+        UniversalBinary = sql.LargeBinary
+    else:
+        UniversalBinary = sql.Binary
+
+import config
 import util
 
 
+# prevent call to socket.getfqdn
+def fast_address_string(self):
+    return '%s' % self.client_address[0]
+BaseHTTPServer.BaseHTTPRequestHandler.address_string = fast_address_string
+del fast_address_string
+
+
+MAX_WORKUNIT_SIZE = int(config.cfg['workunit_size'])
+if MAX_WORKUNIT_SIZE < 1 or MAX_WORKUNIT_SIZE > 1000000:
+    raise ValueError("Invalid 'workunit_size' in configuration")
+
+URL_GROUPER = re.compile("(?P<protocol>\w+)://(((?P<user>\w+):?(?P<passwd>\w+)?@)?(?P<tail>.+))?")
+XMLFAULT = re.compile("\<class '(?P<class>[\w\.]+)'\>:(?P<fault>.+)")
+
+
+def handle_xmlfault(*params):
+    """Decorate a function to check for and rebuild storage exceptions from
+       xmlrpclib.Fault
+    """
+
+    def check_xmlfault(f):
+
+        def protected_f(*args, **kwds):
+            try:
+                ret = f(*args, **kwds)
+            except xmlrpclib.Fault, e:
+                # rpc does not know Exceptions so they always come as pure
+                # strings. One way would be to hack into the de-marshalling.
+                # These seems easier and less intrusive.
+                match = XMLFAULT.match(e.faultString)
+                if match is None:
+                    raise
+                else:
+                    groups = match.groupdict()
+                    cls = groups['class']
+                    fault = groups['fault']
+                    if cls == 'cpyrit.storage.DigestError':
+                        raise DigestError(fault)
+                    elif cls == 'cpyrit.storage.StorageError':
+                        raise StorageError(fault)
+                    else:
+                        raise
+            return ret
+        protected_f.func_name = f.func_name
+        protected_f.__doc__ = f.__doc__
+        return protected_f
+    return check_xmlfault
+
+
+def pruneURL(url):
+    """Remove user/passwd from a storage-url"""
+    match = URL_GROUPER.match(url)
+    if match is None:
+        return url
+    else:
+        url_parts = match.groupdict()
+        protocol = url_parts['protocol']
+        if protocol is None:
+            protocol = ''
+        tail = url_parts['tail']
+        if tail is None:
+            tail = ''
+        return "%s://%s" % (protocol, tail)
+
+
 def getStorage(url):
     if not '://' in url:
         raise ValueError("URL must be of form [protocol]://" \
@@ -68,18 +134,12 @@ def getStorage(url):
         raise RuntimeError("The protocol '%s' is unsupported." % protocol)
 
 
-class ResultCollection(object):
-    """A abstract collection of (Password,PMK)-tuples"""
+class StorageError(IOError):
+    pass
 
-    def __init__(self, essid=None, results=None):
-        self.results = results
-        self.essid = essid
 
-    def __iter__(self):
-        return self.results.__iter__()
-
-    def __len__(self):
-        return len(self.results)
+class DigestError(StorageError):
+    pass
 
 
 class PasswordCollection(object):
@@ -101,54 +161,70 @@ class BasePYR_Buffer(object):
     pyr_head = '<4sH'
     pyr_len = struct.calcsize(pyr_head)
 
+    def __init__(self, essid=None, results=None):
+        self.results = results
+        self.essid = essid
+        self._unpackLock = threading.Lock()
+
     def unpack(self, buf):
-            self._magic, essidlen = struct.unpack(self.pyr_head, buf[:self.pyr_len])
-            if self._magic == 'PYR2':
-                self._delimiter = '\n'
-            elif self._magic == 'PYRT':
-                self._delimiter = '\00'
-            else:
-                raise ValueError("Not a PYRT- or PYR2-buffer.")
-            headfmt = "<%ssi16s" % (essidlen, )
-            headsize = struct.calcsize(headfmt)
-            header = struct.unpack(headfmt, \
-                                   buf[self.pyr_len:self.pyr_len + headsize])
-            self.essid, self._numElems, self._digest = header
-            pmkoffset = self.pyr_len + headsize
-            pwoffset = pmkoffset + self._numElems * 32
-            self._pwbuffer = buf[pwoffset:]
-            self._pmkbuffer = buf[pmkoffset:pwoffset]
-            if len(self._pmkbuffer) % 32 != 0:
-                raise RuntimeError("pmkbuffer seems truncated")
+        if len(buf) < self.pyr_len:
+            raise StorageError("Buffer too short")
+        self._magic, essidlen = struct.unpack(self.pyr_head, \
+                                              buf[:self.pyr_len])
+        if self._magic == 'PYR2':
+            self._delimiter = '\n'
+        elif self._magic == 'PYRT':
+            self._delimiter = '\00'
+        else:
+            raise StorageError("Not a PYRT- or PYR2-buffer.")
+        headfmt = "<%ssi16s" % (essidlen, )
+        headsize = struct.calcsize(headfmt)
+        header = buf[self.pyr_len:self.pyr_len + headsize]
+        if len(header) != headsize:
+            raise StorageError("Invalid header size")
+        header = struct.unpack(headfmt, header)
+        self.essid, self._numElems, self._digest = header
+        pmkoffset = self.pyr_len + headsize
+        pwoffset = pmkoffset + self._numElems * 32
+        self._pwbuffer = buf[pwoffset:]
+        self._pmkbuffer = buf[pmkoffset:pwoffset]
+        if len(self._pmkbuffer) % 32 != 0:
+            raise StorageError("pmkbuffer seems truncated")
 
     def _unpack(self):
-        if hasattr(self, '_pwbuffer'):
-            pwbuffer = zlib.decompress(self._pwbuffer).split(self._delimiter)
-            if len(pwbuffer) != self._numElems:
-                raise RuntimeError("Wrong number of elements")
-            md = hashlib.md5()
-            md.update(self.essid)
-            if self._magic == 'PYR2':
-                md.update(self._pmkbuffer)
-                md.update(self._pwbuffer)
-            else:
-                md.update(self._pmkbuffer)
-                md.update(''.join(pwbuffer))
-            if md.digest() != self._digest:
-                raise IOError("Digest check failed")
-            self.results = zip(pwbuffer, util.grouper(self._pmkbuffer, 32))
-            del self._pwbuffer
-            del self._digest
-            del self._magic
-            del self._delimiter
+        with self._unpackLock:
+            if hasattr(self, '_pwbuffer'):
+                pwbuffer = zlib.decompress(self._pwbuffer)
+                pwbuffer = pwbuffer.split(self._delimiter)
+                assert len(pwbuffer) == self._numElems
+                md = hashlib.md5()
+                md.update(self.essid)
+                if self._magic == 'PYR2':
+                    md.update(self._pmkbuffer)
+                    md.update(self._pwbuffer)
+                else:
+                    md.update(self._pmkbuffer)
+                    md.update(''.join(pwbuffer))
+                if md.digest() != self._digest:
+                    raise DigestError("Digest check failed")
+                self.results = zip(pwbuffer, util.grouper(self._pmkbuffer, 32))
+                assert len(self.results) == self._numElems
+                del self._pwbuffer
+                del self._digest
+                del self._magic
+                del self._delimiter
 
     def __getitem__(self, idx):
         self._unpack()
-        return self.results[idx]
+        try:
+            return self.results[idx]
+        except IndexError:
+            raise IndexError("Instance has %i results, index %i invalid." % \
+                             (len(self.results), idx))
 
     def __len__(self):
         if not hasattr(self, '_numElems'):
-            return len(self.collection)
+            return len(self.results)
         else:
             return self._numElems
 
@@ -160,11 +236,11 @@ class BasePYR_Buffer(object):
         return buffer(self._pmkbuffer)
 
 
-class PYRT_Buffer(ResultCollection, BasePYR_Buffer):
+class PYRT_Buffer(BasePYR_Buffer):
     pass
 
 
-class PYR2_Buffer(BasePYR_Buffer, ResultCollection):
+class PYR2_Buffer(BasePYR_Buffer):
 
     def pack(self):
         pws, pmks = zip(*self.results)
@@ -184,12 +260,12 @@ class PAWD_Buffer(PasswordCollection):
 
     def unpack(self, buf):
         if buf[:4] != "PAWD":
-            raise ValueError("Not a PAWD-buffer.")
+            raise StorageError("Not a PAWD-buffer.")
         md = hashlib.md5()
         inp = tuple(buf[4 + md.digest_size:].split('\00'))
         md.update(''.join(inp))
         if buf[4:4 + md.digest_size] != md.digest():
-            raise IOError("Digest check failed.")
+            raise DigestError("Digest check failed.")
         self.collection = inp
         self.key = md.hexdigest()
 
@@ -204,11 +280,11 @@ class PAW2_Buffer(PasswordCollection):
 
     def unpack(self, buf):
         if buf[:4] != "PAW2":
-            raise ValueError("Not a PAW2-buffer.")
+            raise StorageError("Not a PAW2-buffer.")
         md = hashlib.md5()
         md.update(buf[4 + md.digest_size:])
         if md.digest() != buf[4:4 + md.digest_size]:
-            raise IOError("Digest check failed.")
+            raise DigestError("Digest check failed")
         inp = tuple(zlib.decompress(buf[4 + md.digest_size:]).split('\n'))
         self.collection = inp
         self.key = md.hexdigest()
@@ -273,7 +349,7 @@ class PasswordStore(object):
            performance and efficiency. It is the caller's responsibility to
            call .flush_buffer() (or use the context-manager) when he is done.
         """
-        passwd = passwd.strip()
+        passwd = passwd.strip('\r\n')
         if len(passwd) < 8 or len(passwd) > 63:
             return
         pw_h1 = PasswordStore.h1_list[hash(passwd) & 0xFF]
@@ -283,7 +359,7 @@ class PasswordStore(object):
         else:
             pw_bucket = self.pwbuffer.setdefault(pw_h1, list())
             pw_bucket.append(passwd)
-        if len(pw_bucket) >= 20000:
+        if len(pw_bucket) >= MAX_WORKUNIT_SIZE:
             self._flush_bucket(pw_h1, pw_bucket)
             self.pwbuffer[pw_h1] = (set if self.unique_check else list)()
 
@@ -304,6 +380,12 @@ class PasswordStore(object):
 
 class Storage(object):
 
+    def __delitem__(self, key):
+        for essid in self.essids:
+            if self.essids.containskey(essid, key):
+                del self.essids[essid, key]
+        del self.passwords[key]
+
     def iterresults(self, essid):
         return self.essids.iterresults(essid)
 
@@ -363,10 +445,10 @@ class FSEssidStore(ESSIDStore):
                 self.essids[essid] = (essidpath, {})
                 for pyrfile in os.listdir(essidpath):
                     if pyrfile.endswith('.pyr'):
-                        self.essids[essid][1][pyrfile[:len(pyrfile)-4]] = \
+                        self.essids[essid][1][pyrfile[:len(pyrfile) - 4]] = \
                                             os.path.join(essidpath, pyrfile)
             else:
-                print >> sys.stderr, "ESSID %s is corrupted." % essid_hash
+                print >>sys.stderr, "ESSID %s is corrupted." % essid_hash
 
     def __getitem__(self, (essid, key)):
         """Receive a iterable of (password,PMK)-tuples stored under
@@ -375,25 +457,23 @@ class FSEssidStore(ESSIDStore):
            Returns a empty iterable if the key is not stored. Raises KeyError
            if the ESSID is not stored.
         """
-        if not self.containskey(essid, key):
-            return ()
         try:
-            with open(self.essids[essid][1][key], 'rb') as f:
+            fname = self.essids[essid][1][key]
+        except IndexError:
+            raise IndexError("No result for ESSID:Key (%s:%s)" % (essid, key))
+        else:
+            with open(fname, 'rb') as f:
                 buf = f.read()
             if buf.startswith('PYR2'):
                 results = PYR2_Buffer()
             elif buf.startswith('PYRT'):
                 results = PYRT_Buffer()
             else:
-                raise IOError("File-format for '%s' unknown." % filename)
+                raise StorageError("Header for '%s' unknown." % (fname,))
             results.unpack(buf)
             if results.essid != essid:
-                raise RuntimeError("Invalid ESSID in result-collection")
+                raise StorageError("Invalid ESSID in result-collection")
             return results
-        except:
-            print >> sys.stderr, "Error while loading results %s for " \
-                                 "ESSID '%s'" % (key, essid)
-            raise
 
     def __setitem__(self, (essid, key), results):
         """Store a iterable of (password,PMK)-tuples under the given
@@ -418,16 +498,25 @@ class FSEssidStore(ESSIDStore):
         """Return True if the given ESSID is currently stored."""
         return essid in self.essids
 
-    def __delitem__(self, essid):
-        """Delete the given ESSID and all results from the storage."""
+    def __delitem__(self, (essid, key)):
+        """Delete the given ESSID:key resultset or the entire ESSID
+           and all results from the storage.
+        """
         if essid not in self:
-            raise KeyError("ESSID not in store.")
-        essid_root, pyrfiles = self.essids[essid]
-        del self.essids[essid]
-        for fname in pyrfiles.itervalues():
+            raise KeyError("ESSID not in storage")
+        if key is not None:
+            if key not in self.essids[essid][1]:
+                raise KeyError("Key not in storage")
+            essid_root, pyrfiles = self.essids[essid]
+            fname = pyrfiles.pop(key)
             os.unlink(fname)
-        os.unlink(os.path.join(essid_root, 'essid'))
-        os.rmdir(essid_root)
+        else:
+            essid_root, pyrfiles = self.essids[essid]
+            del self.essids[essid]
+            for fname in pyrfiles.itervalues():
+                os.unlink(fname)
+            os.unlink(os.path.join(essid_root, 'essid'))
+            os.rmdir(essid_root)
 
     def containskey(self, essid, key):
         """Return True if the given (ESSID,key) combination is stored."""
@@ -455,7 +544,7 @@ class FSEssidStore(ESSIDStore):
            Re-creating a ESSID is a no-op.
         """
         if len(essid) < 1 or len(essid) > 32:
-            raise ValueError("ESSID invalid.")
+            raise ValueError("ESSID '%s' invalid." % (essid, ))
         root = os.path.join(self.basepath, hashlib.md5(essid).hexdigest()[:8])
         if not os.path.exists(root):
             os.makedirs(root)
@@ -479,7 +568,7 @@ class FSPasswordStore(PasswordStore):
             for pwfile in os.listdir(pwpath):
                 if pwfile[-3:] != '.pw':
                     continue
-                self.pwfiles[pwfile[:len(pwfile)-3]] = pwpath
+                self.pwfiles[pwfile[:len(pwfile) - 3]] = pwpath
 
     def __contains__(self, key):
         """Return True if the given key is currently in the storage."""
@@ -498,23 +587,25 @@ class FSPasswordStore(PasswordStore):
     def __getitem__(self, key):
         """Return the collection of passwords indexed by the given key."""
         filename = os.path.join(self.pwfiles[key], key) + '.pw'
-        try:
-            with open(filename, 'rb') as f:
-                buf = f.read()
-            if buf[:4] == "PAW2":
-                inp = PAW2_Buffer()
-            elif buf[:4] == "PAWD":
-                inp = PAWD_Buffer()
-            else:
-                raise IOError("'%s' is not a PasswordFile." % filename)
-            inp.unpack(buf)
-            if inp.key != key:
-                raise IOError("File doesn't match the key '%s'." % inp.key)
-        except:
-            print >> sys.stdout, "Error while opening '%s'" % filename
-            raise
+        with open(filename, 'rb') as f:
+            buf = f.read()
+        if buf[:4] == "PAW2":
+            inp = PAW2_Buffer()
+        elif buf[:4] == "PAWD":
+            inp = PAWD_Buffer()
+        else:
+            raise StorageError("'%s' is not a PasswordFile." % filename)
+        inp.unpack(buf)
+        if inp.key != key:
+            raise StorageError("File doesn't match the key '%s'." % inp.key)
         return inp
 
+    def __delitem__(self, key):
+        """Delete workunit from storage"""
+        path = self.pwfiles.pop(key)
+        filename = os.path.join(path, key) + '.pw'
+        os.unlink(filename)
+
     def size(self, key):
         """Return the number of passwords indexed by the given key."""
         return len(self[key])
@@ -544,6 +635,7 @@ class RPCStorage(Storage):
         self.essids = RPCESSIDStore(self.cli)
         self.passwords = RPCPasswordStore(self.cli)
 
+    @handle_xmlfault()
     def getStats(self):
         return self.cli.getStats()
 
@@ -553,6 +645,7 @@ class RPCESSIDStore(ESSIDStore):
     def __init__(self, cli):
         self.cli = cli
 
+    @handle_xmlfault()
     def __getitem__(self, (essid, key)):
         """Receive a iterable of (password,PMK)-tuples stored under
            the given ESSID and key.
@@ -561,12 +654,16 @@ class RPCESSIDStore(ESSIDStore):
            if the ESSID is not stored.
         """
         buf = self.cli.essids.getitem(essid, key)
-        results = PYR2_Buffer()
-        results.unpack(buf.data)
-        if results.essid != essid:
-            raise ValueError("Invalid ESSID in result-collection")
-        return results
+        if buf:
+            results = PYR2_Buffer()
+            results.unpack(buf.data)
+            if results.essid != essid:
+                raise StorageError("Invalid ESSID in result-collection")
+            return results
+        else:
+            raise KeyError
 
+    @handle_xmlfault()
     def __setitem__(self, (essid, key), results):
         """Store a iterable of (password,PMK)-tuples under the given
            ESSID and key.
@@ -574,36 +671,49 @@ class RPCESSIDStore(ESSIDStore):
         b = xmlrpclib.Binary(PYR2_Buffer(essid, results).pack())
         self.cli.essids.setitem(essid, key, b)
 
+    @handle_xmlfault()
     def __len__(self):
         """Return the number of ESSIDs currently stored."""
         return self.cli.essids.len()
 
+    @handle_xmlfault()
     def __iter__(self):
         """Iterate over all essids currently stored."""
         return self.cli.essids.essids().__iter__()
 
+    @handle_xmlfault()
     def __contains__(self, essid):
         """Return True if the given ESSID is currently stored."""
         return self.cli.essids.contains(essid)
 
-    def __delitem__(self, essid):
-        """Delete the given ESSID and all results from the storage."""
-        self.cli.essids.delitem(essid)
+    @handle_xmlfault()
+    def __delitem__(self, (essid, key)):
+        """Delete the ESSID:key resultset or the entire ESSID
+           and all results from the storage.
+        """
+        if key is None:
+            key = ''
+        self.cli.essids.delitem(essid, key)
+        return True
 
+    @handle_xmlfault()
     def containskey(self, essid, key):
         """Return True if the given (ESSID,key) combination is stored."""
         return self.cli.essids.containskey(essid, key)
 
+    @handle_xmlfault()
     def keycount(self, essid):
         """Returns the number of keys that can currently be used to receive
            results for the given ESSID.
         """
         return self.cli.essids.keycount(essid)
 
+    @handle_xmlfault()
     def iterkeys(self, essid):
         """Iterate over all keys that can be used to receive results."""
         return self.cli.essids.keys(essid).__iter__()
 
+    @handle_xmlfault()
     def create_essid(self, essid):
         """Create the given ESSID in the storage.
 
@@ -615,64 +725,73 @@ class RPCESSIDStore(ESSIDStore):
 class RPCPasswordStore(PasswordStore):
 
     def __init__(self, cli):
+        PasswordStore.__init__(self)
         self.cli = cli
 
+    @handle_xmlfault()
     def __iter__(self):
         """Iterate over all keys that can be used to receive password-sets."""
         return self.cli.passwords.keys().__iter__()
 
+    @handle_xmlfault()
     def __len__(self):
         """Return the number of keys that can be used to receive
            password-sets.
         """
         return self.cli.passwords.len()
 
+    @handle_xmlfault()
     def __contains__(self, key):
         """Return True if the given key is currently in the storage."""
         return self.cli.passwords.contains(key)
 
+    @handle_xmlfault()
     def __getitem__(self, key):
         """Return the collection of passwords indexed by the given key."""
         bucket = PAW2_Buffer()
         bucket.unpack(self.cli.passwords.getitem(key).data)
         if bucket.key != key:
-            raise IOError("File doesn't match the key '%s'." % bucket.key)
+            raise StorageError("File doesn't match the key '%s'." % bucket.key)
         return bucket
 
+    @handle_xmlfault()
+    def __delitem__(self, key):
+        return self.cli.passwords.delitem(key)
+
+    @handle_xmlfault()
     def size(self, key):
         """Return the number of passwords indexed by the given key."""
         return self.cli.passwords.size(key)
 
+    @handle_xmlfault()
+    def _flush_bucket(self, pw_h1, bucket):
+        return self.cli.passwords._flush_bucket(pw_h1, tuple(bucket))
 
-class RPCServer(SimpleXMLRPCServer.SimpleXMLRPCServer):
 
-    def __init__(self, storage):
-        SimpleXMLRPCServer.SimpleXMLRPCServer.__init__(self, ('', 17934), \
-                                                        logRequests=False)
+class StorageRelay(util.AsyncXMLRPCServer):
+
+    def __init__(self, storage, iface='', port=17934):
+        util.AsyncXMLRPCServer.__init__(self, (iface, port))
+        self.methods['getStats'] = self.getStats
+        self.methods['passwords.keys'] = self.passwords_keys
+        self.methods['passwords.len'] = self.passwords_len
+        self.methods['passwords.contains'] = self.passwords_contains
+        self.methods['passwords.getitem'] = self.passwords_getitem
+        self.methods['passwords.delitem'] = self.passwords_delitem
+        self.methods['passwords.size'] = self.passwords_size
+        self.methods['passwords._flush_bucket'] = self.passwords_flush
+        self.methods['essids.essids'] = self.essids_essids
+        self.methods['essids.keys'] = self.essids_keys
+        self.methods['essids.len'] = self.essids_len
+        self.methods['essids.contains'] = self.essids_contains
+        self.methods['essids.getitem'] = self.essids_getitem
+        self.methods['essids.setitem'] = self.essids_setitem
+        self.methods['essids.delitem'] = self.essids_delitem
+        self.methods['essids.keycount'] = self.essids_keycount
+        self.methods['essids.containskey'] = self.essids_containskey
+        self.methods['essids.createessid'] = self.essids_create_essid
         self.storage = storage
-        self.methods = {'getStats': self.getStats, \
-                        'passwords.keys': self.passwords_keys, \
-                        'passwords.len': self.passwords_len, \
-                        'passwords.contains': self.passwords_contains, \
-                        'passwords.getitem': self.passwords_getitem, \
-                        'passwords.size': self.passwords_size, \
-                        'essids.essids': self.essids_essids, \
-                        'essids.keys': self.essids_keys, \
-                        'essids.len': self.essids_len, \
-                        'essids.contains': self.essids_contains, \
-                        'essids.getitem': self.essids_getitem, \
-                        'essids.setitem': self.essids_setitem, \
-                        'essids.delitem': self.essids_delitem, \
-                        'essids.keycount': self.essids_keycount, \
-                        'essids.containskey': self.essids_containskey, \
-                        'essids.createessid': self.essids_create_essid}
-        self.register_instance(self)
-
-    def _dispatch(self, method, params):
-        if method not in self.methods:
-            raise AttributeError
-        else:
-            return self.methods[method](*params)
+        self.start()
 
     def getStats(self):
         return self.storage.getStats()
@@ -690,9 +809,17 @@ class RPCServer(SimpleXMLRPCServer.SimpleXMLRPCServer):
         newkey, buf = PAW2_Buffer(self.storage.passwords[key]).pack()
         return xmlrpclib.Binary(buf)
 
+    def passwords_delitem(self, key):
+        del self.storage.passwords[key]
+        return True
+
     def passwords_size(self, key):
         return self.storage.passwords.size(key)
 
+    def passwords_flush(self, pw_h1, bucket):
+        self.storage.passwords._flush_bucket(pw_h1, set(bucket))
+        return True
+
     def essids_essids(self):
         return list(self.storage.essids)
 
@@ -706,9 +833,13 @@ class RPCServer(SimpleXMLRPCServer.SimpleXMLRPCServer):
         return essid in self.storage.essids
 
     def essids_getitem(self, essid, key):
-        results = self.storage.essids[essid, key]
-        buf = PYR2_Buffer(essid, results).pack()
-        return xmlrpclib.Binary(buf)
+        try:
+            results = self.storage.essids[essid, key]
+        except KeyError:
+            return False
+        else:
+            buf = PYR2_Buffer(essid, results).pack()
+            return xmlrpclib.Binary(buf)
 
     def essids_setitem(self, essid, key, pyr2_buffer):
         results = PYR2_Buffer()
@@ -718,8 +849,10 @@ class RPCServer(SimpleXMLRPCServer.SimpleXMLRPCServer):
         self.storage.essids[essid, key] = results
         return True
 
-    def essids_delitem(self, essid):
-        del self.storage.essids[essid]
+    def essids_delitem(self, essid, key=None):
+        if key == '':
+            key = None
+        del self.storage.essids[essid, key]
         return True
 
     def essids_keycount(self, essid):
@@ -737,37 +870,38 @@ if 'sqlalchemy' in sys.modules:
     metadata = sql.MetaData()
 
     essids_table = sql.Table('essids', metadata, \
-                        sql.Column('essid_id', sql.Integer, primary_key=True),
-                        sql.Column('essid', sql.Binary(32), nullable=False),
-                        sql.Column('uid', sql.String(32), unique=True, \
-                                    nullable=False), \
-                        mysql_engine='InnoDB')
+                    sql.Column('essid_id', sql.Integer, primary_key=True),
+                    sql.Column('essid', UniversalBinary(32), nullable=False),
+                    sql.Column('uid', sql.String(32), unique=True, \
+                                nullable=False), \
+                    mysql_engine='InnoDB')
 
     passwords_table = sql.Table('passwords', metadata, \
-                        sql.Column('_key', sql.String(32), primary_key=True),
-                        sql.Column('h1', sql.String(2), nullable=False),
-                        sql.Column('numElems', sql.Integer, nullable=False),
-                        sql.Column('collection_buffer', sql.Binary(2**24-1), \
-                                   nullable=False), \
-                        mysql_engine='InnoDB')
+                    sql.Column('_key', sql.String(32), primary_key=True),
+                    sql.Column('h1', sql.String(2), nullable=False, \
+                               index=True),
+                    sql.Column('numElems', sql.Integer, nullable=False),
+                    sql.Column('collection_buffer', UniversalBinary(2**24-1), \
+                               nullable=False), \
+                    mysql_engine='InnoDB')
 
     results_table = sql.Table('results', metadata, \
-                        sql.Column('_key', sql.String(32), \
-                                   sql.ForeignKey('passwords._key'), \
-                                   primary_key=True),
-                        sql.Column('essid_id', sql.Integer, \
-                                   sql.ForeignKey('essids.essid_id'), \
-                                   primary_key=True),
-                        sql.Column('numElems', sql.Integer, nullable=False),
-                        sql.Column('results_buffer', sql.Binary(2**24 - 1), \
-                                   nullable=False), \
-                        mysql_engine='InnoDB')
+                    sql.Column('essid_id', sql.Integer, \
+                               sql.ForeignKey('essids.essid_id'), \
+                               primary_key=True),
+                    sql.Column('_key', sql.String(32), \
+                               sql.ForeignKey('passwords._key'), \
+                               primary_key=True, index=True),
+                    sql.Column('numElems', sql.Integer, nullable=False),
+                    sql.Column('results_buffer', UniversalBinary(2**24 - 1), \
+                               nullable=False), \
+                    mysql_engine='InnoDB')
 
     class ESSID_DBObject(object):
 
         def __init__(self, essid):
             if len(essid) < 1 or len(essid) > 32:
-                raise ValueError("ESSID invalid")
+                raise ValueError("ESSID '%s' invalid" % (essid, ))
             self.essid = essid
             self.uid = hashlib.md5(essid).hexdigest()
 
@@ -859,9 +993,9 @@ if 'sqlalchemy' in sys.modules:
     class SQLStorage(Storage):
 
         def __init__(self, url):
-            engine = sql.create_engine(url, echo=False)
-            metadata.create_all(engine)
-            self.SessionClass = orm.sessionmaker(bind=engine)
+            self.engine = sql.create_engine(url, echo=False)
+            metadata.create_all(self.engine)
+            self.SessionClass = orm.sessionmaker(bind=self.engine)
             self.essids = SQLEssidStore(self.SessionClass)
             self.passwords = SQLPasswordStore(self.SessionClass)
 
@@ -917,7 +1051,8 @@ if 'sqlalchemy' in sys.modules:
                 result = q.filter(sql.and_(ESSID_DBObject.essid == essid, \
                                            PYR2_DBObject.key == key)).first()
                 if result is None:
-                    return ()
+                    raise KeyError("No result for ESSID:Key-combination " \
+                                     "(%s:%s)." % (essid, key))
                 else:
                     return result
 
@@ -928,45 +1063,38 @@ if 'sqlalchemy' in sys.modules:
             with SessionContext(self.SessionClass) as session:
                 q = session.query(ESSID_DBObject)
                 essid_obj = q.filter(ESSID_DBObject.essid == essid).one()
-                q = session.query(PYR2_DBObject).join(ESSID_DBObject)
-                q = q.filter(sql.and_( \
-                                    ESSID_DBObject.essid == essid_obj.essid, \
-                                    PYR2_DBObject.key == key))
-                result_obj = q.first()
-                if result_obj is None:
-                    session.add(PYR2_DBObject(essid_obj, key, results))
-                    try:
-                        session.commit()
-                    except sql.exc.IntegrityError:
-                        # Assume we hit a concurrent insert that causes
-                        # a constraint-error on (essid-key).
-                        session.rollback()
-                        q = session.query(PYR2_DBObject).join(ESSID_DBObject)
-                        q = q.filter(sql.and_( \
-                                     ESSID_DBObject.essid == essid_obj.essid, \
-                                     PYR2_DBObject.key == key))
-                        result_obj = q.one()
-                        result_obj.pack(results)
-                        session.commit()
-                else:
+                session.add(PYR2_DBObject(essid_obj, key, results))
+                try:
+                    session.commit()
+                except sql.exc.IntegrityError:
+                    # Assume we hit a concurrent insert that causes
+                    # a constraint-error on (essid-key).
+                    session.rollback()
+                    q = session.query(PYR2_DBObject).join(ESSID_DBObject)
+                    q = q.filter(sql.and_( \
+                                 ESSID_DBObject.essid == essid_obj.essid, \
+                                 PYR2_DBObject.key == key))
+                    result_obj = q.one()
                     result_obj.pack(results)
                     session.commit()
 
-        def __delitem__(self, essid):
-            """Delete the given ESSID and all results from the storage."""
+        def __delitem__(self, (essid, key)):
+            """Delete the given ESSID:key resultset or the entire ESSID
+               and all results from the storage.
+            """
             with SessionContext(self.SessionClass) as session:
                 essid_query = session.query(ESSID_DBObject)
                 essid_query = essid_query.filter(ESSID_DBObject.essid == essid)
                 essid_obj = essid_query.one()
                 res_query = session.query(PYR2_DBObject)
                 res_query = res_query.filter(PYR2_DBObject.essid == essid_obj)
-                res_query.delete(synchronize_session=False)
-                cnt = essid_query.delete(synchronize_session=False)
-                if cnt > 1:
-                    raise ValueError("Query should not have affected more " \
-                                     "than one row.")
+                if key is not None:
+                    res_query = res_query.filter(PYR2_DBObject.key == key)
+                    assert res_query.delete(synchronize_session=False) == 1
                 else:
-                    session.commit()
+                    res_query.delete(synchronize_session=False)
+                    assert essid_query.delete(synchronize_session=False) == 1
+                session.commit()
 
         def containskey(self, essid, key):
             """Return True if the given (ESSID,key) combination is stored."""
@@ -1044,6 +1172,14 @@ if 'sqlalchemy' in sys.modules:
                 q = session.query(PAW2_DBObject)
                 return q.filter(PAW2_DBObject.key == key).one()
 
+        def __delitem__(self, key):
+            """Delete the collection of passwords indexed by the given key."""
+            with SessionContext(self.SessionClass) as session:
+                q = session.query(PAW2_DBObject)
+                q = q.filter(PAW2_DBObject.key == key)
+                assert q.delete(synchronize_session=False) == 1
+                session.commit()
+
         def size(self, key):
             """Return the number of passwords indexed by the given key."""
             with SessionContext(self.SessionClass) as session:
