@@ -1,6 +1,6 @@
 # -*- coding: UTF-8 -*-
 #
-#    Copyright 2008-2010, Lukas Lueg, lukas.lueg@gmail.com
+#    Copyright 2008-2011, Lukas Lueg, lukas.lueg@gmail.com
 #
 #    This file is part of Pyrit.
 #
@@ -17,18 +17,20 @@
 #    You should have received a copy of the GNU General Public License
 #    along with Pyrit.  If not, see <http://www.gnu.org/licenses/>.
 
-"""Various utility- and backend- related classes and data for Pyrit.
+"""Various utility- and backend-related classes and data for Pyrit.
 
    AsyncFileWriter is used for threaded, buffered output.
 
-   StorageIterator and PassthroughIterator encapsulate the repetitive task of
-   getting workunits from the database, passing them to the hardware if
-   necessary and yielding the results to a client.
-
    CowpattyFile eases reading/writing files in cowpatty's binary format.
 
    ncpus equals the number of available CPUs in the system.
 
+   Thread is a subclass of threading.Thread that adds a context-manager to
+   make it 'stoppable'.
+
+   AsyncXMLRPCServer is a stoppable (incl. 'serve_forever') subclass of
+   SimpleXMLRPCServer.
+
    PMK_TESTVECTORS has two ESSIDs and ten password:PMK pairs each to verify
    local installations.
 """
@@ -40,6 +42,8 @@ import gzip
 import os
 import Queue
 import random
+import socket
+import SimpleXMLRPCServer
 import sys
 import struct
 import time
@@ -47,8 +51,9 @@ import threading
 
 import _cpyrit_cpu
 from _cpyrit_cpu import VERSION, grouper
-import cpyrit
-import storage
+import config
+
+__version__ = VERSION
 
 
 def _detect_ncpus():
@@ -72,7 +77,22 @@ def _detect_ncpus():
     #return the default value
     return 1
 
-ncpus = _detect_ncpus()
+
+def _limit_ncpus():
+    """Limit the number of reported CPUs if so requested"""
+    detected_ncpus = _detect_ncpus()
+    try:
+        limited_ncpus = int(config.cfg['limit_ncpus'])
+    except ValueError:
+        raise ValueError("Invalid 'limit_ncpus' in configuration")
+    if limited_ncpus < 0:
+        raise ValueError("Invalid 'limit_ncpus' in configuration")
+    if limited_ncpus > 0 and limited_ncpus < detected_ncpus:
+        return limited_ncpus
+    return detected_ncpus
+
+
+ncpus = _limit_ncpus()
 """ Number of effective CPUs (in the moment the module was loaded)."""
 
 
@@ -94,111 +114,6 @@ class SqlalchemyImportError(ImportError):
     pass
 
 
-class StorageIterator(object):
-    """Iterates over the database, computes new Pairwise Master Keys if
-       necessary and requested and yields tuples of (password,PMK)-tuples.
-    """
-
-    def __init__(self, storage, essid, \
-                 yieldOldResults=True, yieldNewResults=True):
-        self.cp = None
-        self.workunits = []
-        self.essid = essid
-        self.storage = storage
-        keys = list(self.storage.passwords)
-        random.shuffle(keys)
-        self.keys = iter(keys)
-        self.yieldOldResults = yieldOldResults
-        self.yieldNewResults = yieldNewResults
-
-    def keycount(self):
-        return self.storage.essids.keycount(self.essid)
-
-    def __len__(self):
-        return len(self.storage.passwords)
-
-    def __iter__(self):
-        return self
-
-    def __enter__(self):
-        return self
-
-    def __exit__(self, exc_type, exc_val, exc_tb):
-        if self.cp is not None:
-            self.cp.shutdown()
-
-    def next(self):
-        for key in self.keys:
-            if self.storage.essids.containskey(self.essid, key):
-                if self.yieldOldResults:
-                    return self.storage.essids[self.essid, key]
-            else:
-                if self.yieldNewResults:
-                    if self.cp is None:
-                        self.cp = cpyrit.CPyrit()
-                    passwords = self.storage.passwords[key]
-                    self.workunits.append((self.essid, key, passwords))
-                    self.cp.enqueue(self.essid, passwords)
-                    solvedPMKs = self.cp.dequeue(block=False)
-                    if solvedPMKs is not None:
-                        solvedEssid, solvedKey, solvedPasswords = \
-                            self.workunits.pop(0)
-                        solvedResults = zip(solvedPasswords, solvedPMKs)
-                        self.storage.essids[solvedEssid, solvedKey] = \
-                            solvedResults
-                        return solvedResults
-        if self.yieldNewResults and self.cp is not None:
-            for solvedPMKs in self.cp:
-                solvedEssid, solvedKey, solvedPasswords = self.workunits.pop(0)
-                solvedResults = zip(solvedPasswords, solvedPMKs)
-                self.storage.essids[solvedEssid, solvedKey] = solvedResults
-                return solvedResults
-        raise StopIteration
-
-
-class PassthroughIterator(object):
-    """A iterator that takes an ESSID and an iterable of passwords, computes
-       the corresponding Pairwise Master Keys and and yields tuples of
-       (password,PMK)-tuples.
-    """
-
-    def __init__(self, essid, iterable, buffersize=20000):
-        self.cp = cpyrit.CPyrit()
-        self.essid = essid
-        self.iterator = iter(iterable)
-        self.workunits = []
-        self.buffersize = buffersize
-
-    def __iter__(self):
-        return self
-
-    def __enter__(self):
-        return self
-
-    def __exit__(self, exc_type, exc_val, exc_tb):
-        self.cp.shutdown()
-
-    def next(self):
-        pwbuffer = []
-        for line in self.iterator:
-            pw = line.strip()[:63]
-            if len(pw) >= 8:
-                pwbuffer.append(pw)
-            if len(pwbuffer) > self.buffersize:
-                self.workunits.append(pwbuffer)
-                self.cp.enqueue(self.essid, self.workunits[-1])
-                pwbuffer = []
-                solvedPMKs = self.cp.dequeue(block=False)
-                if solvedPMKs is not None:
-                    return zip(self.workunits.pop(0), solvedPMKs)
-        if len(pwbuffer) > 0:
-            self.workunits.append(pwbuffer)
-            self.cp.enqueue(self.essid, self.workunits[-1])
-        for solvedPMKs in self.cp:
-            return zip(self.workunits.pop(0), solvedPMKs)
-        raise StopIteration
-
-
 class FileWrapper(object):
     """A wrapper for easy stdin/stdout/gzip-handling"""
 
@@ -320,7 +235,7 @@ class AsyncFileWriter(threading.Thread):
        thread.
     """
 
-    def __init__(self, f, maxsize=10 * 1024**2):
+    def __init__(self, f, maxsize=10 * 1024 ** 2):
         """Create a instance writing to the given file-like-object and
            buffering maxsize before blocking.
         """
@@ -434,7 +349,7 @@ class AsyncFileWriter(threading.Thread):
 
 class PerformanceCounter(object):
 
-    def __init__(self, window=30.0):
+    def __init__(self, window=60.0):
         self.window = window
         self.datapoints = [[time.time(), 0.0]]
         self.total = 0
@@ -460,7 +375,8 @@ class PerformanceCounter(object):
     def __purge(self):
         t = time.time()
         if t - self.datapoints[0][0] > self.window:
-            self.datapoints = filter(lambda x: (t - x[0]) < self.window, self.datapoints)
+            self.datapoints = filter(lambda x: (t - x[0]) < self.window, \
+                                                self.datapoints)
 
     def getAvg(self):
         self.__purge()
@@ -478,6 +394,74 @@ class PerformanceCounter(object):
     avg = property(fget=getAvg)
 
 
+class Thread(threading.Thread):
+    """A stoppable subclass of threading.Thread"""
+
+    def __init__(self):
+        threading.Thread.__init__(self)
+        self.shallStop = False
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.shutdown()
+
+    def shutdown(self):
+        self.shallStop = True
+        self.join()
+
+
+class AsyncXMLRPCServer(SimpleXMLRPCServer.SimpleXMLRPCServer, Thread):
+    """A stoppable XMLRPCServer
+
+       The main socket is made non-blocking so we can check on
+       self.shallStop from time to time.
+
+       Sub-classes should add (name:function)-entries to self.methods
+    """
+
+    def __init__(self, (iface, port)=('', 17934)):
+        SimpleXMLRPCServer.SimpleXMLRPCServer.__init__(self, (iface, port), \
+                                                        logRequests=False)
+        Thread.__init__(self)
+        self.setDaemon(True)
+        # Make the main socket non-blocking (for accept())
+        self.socket.settimeout(1)
+        self.methods = {}
+        self.register_instance(self)
+
+    def run(self):
+        while not self.shallStop:
+            self.handle_request()
+
+    def get_request(self):
+        while not self.shallStop:
+            try:
+                sock, addr = self.socket.accept()
+            except socket.timeout:
+                pass
+            else:
+                # Accepted connections are made blocking again
+                sock.settimeout(None)
+                return sock, addr
+        raise socket.timeout("Server has stopped.")
+
+    def serve_forever(self):
+        while not self.shallStop:
+            time.sleep(1)
+
+    def shutdown(self):
+        Thread.shutdown(self)
+        self.socket.close()
+
+    def _dispatch(self, method, params):
+        if method not in self.methods:
+            raise AttributeError
+        else:
+            return self.methods[method](*params)
+
+
 PMK_TESTVECTORS = {
     'foo': {
         'soZcEvntHVrGRDIxNaBCyUL':
